import java.io.*;
import java.util.*;

/** LexicalAnalyzer implementation.
 *  uses a LineNumberReader to scan through the InputStream.
 */
public class DefaultLexicalAnalyzer implements LexicalAnalyzer {
	
	/** This is the maximum size of pushback characters. */
	public final static int PUSH_BACK_SIZE = 80;
	
	/** Reader for the program InputStream. */
	protected LineNumberReader reader;
	
	/** Tokenizes the Stream according to language definition.
	 *  @param is The InputStream.
	 *  @return List of Tokens.
	 *  @throws IOException if an IO failed.
	 *  @throws LexicalException if the stream contains an undefined character.
	 */
	public List tokenize(InputStream is) throws IOException, LexicalException {
		
		// reader for the InputStream.
		reader = new LineNumberReader(
		         new InputStreamReader(is), PUSH_BACK_SIZE);
				 
		// stores the tokens
		LinkedList tokenList = new LinkedList();

		//  the currently scanned character. 
		int c = 0;
		
		//  stores the token.
		TokenFactory.Token token = null;
		
		// loop through all available chars of the reader.
		do {
			
			// mark the current position so we can unread.
			reader.mark(PUSH_BACK_SIZE);
			c = reader.read();
			
			
			if (c == -1) { // No more chars to read.
				token = TokenFactory.createEOT(reader.getLineNumber());
			} else
			if (c == '+') { // check for aggregations starting with '+'
				reader.mark(1);
				c = reader.read();
				if (c == '+') {
					token = TokenFactory.createToken("++", reader.getLineNumber());
				} else
				if (c == '=') {
					token = TokenFactory.createToken("+=", reader.getLineNumber());
				} else {
					reader.reset();
					token = TokenFactory.createToken("+", reader.getLineNumber());
				}
			} else 
			if (c == '-') { // check for aggregations starting with '-'
				reader.mark(1);
				c = reader.read();
				if (c == '-') {
					token = TokenFactory.createToken("--", reader.getLineNumber());
				} else
				if (c == '=') {
					token = TokenFactory.createToken("-=", reader.getLineNumber());
				} else {
					reader.reset();
					token = TokenFactory.createToken("-", reader.getLineNumber());
				}
			} else 
			if (c == '>') { // no more comments on these checkings ...
				reader.mark(1);
				c = reader.read();
				if (c == '=') {
					token = TokenFactory.createToken(">=", reader.getLineNumber());
				} else {
					reader.reset();
					token = TokenFactory.createToken(">", reader.getLineNumber());
				}
			} else 
			if (c == '<') {
				reader.mark(1);
				c = reader.read();
				if (c == '=') {
					token = TokenFactory.createToken("<=", reader.getLineNumber());
				} else {
					reader.reset();
					token = TokenFactory.createToken("<", reader.getLineNumber());
				}
			} else 
			if (c == '=') {
				reader.mark(1);
				c = reader.read();
				if (c == '=') {
					token = TokenFactory.createToken("==", reader.getLineNumber());
				} else {
					reader.reset();
					token = TokenFactory.createToken("=", reader.getLineNumber());
				}
			} else 
			if (c == '!') {
				reader.mark(1);
				c = reader.read();
				if (c == '=') {
					token = TokenFactory.createToken("!=", reader.getLineNumber());
				} else {
					reader.reset();
					token = TokenFactory.createToken("!", reader.getLineNumber());
				}
			} else 
			if (c == '*') {
				token = TokenFactory.createToken("*", reader.getLineNumber());
			} else 
			if (c == '/') {
				token = TokenFactory.createToken("/", reader.getLineNumber());
			} else 
			if (c == ':') {
				token = TokenFactory.createToken(":", reader.getLineNumber());
			} else 
			if (c == ',') {
				token = TokenFactory.createToken(",", reader.getLineNumber());
			} else 
			if (c == ';') {
				token = TokenFactory.createToken(";", reader.getLineNumber());
			} else 
			if (c == '(') {
				token = TokenFactory.createToken("(", reader.getLineNumber());
			} else 
			if (c == ')') {
				token = TokenFactory.createToken(")", reader.getLineNumber());
			} else 
			if (c == '{') {
				token = TokenFactory.createToken("{", reader.getLineNumber());
			} else 
			if (c == '}') {
				token = TokenFactory.createToken("}", reader.getLineNumber());
			} else 
			if (Character.isDigit((char) c)) { // check for integer literals
				token = TokenFactory.createToken(scanIntegerLiteral(), reader.getLineNumber());
			} else
			if (Character.isJavaIdentifierStart((char) c)) { // check for keywords, identifiers
				token = TokenFactory.createToken(scanIdentifier(), reader.getLineNumber());
			} else 
			if (Character.isWhitespace((char) c)) {
				// skip whitespace
				token = null;
			} else {
				// unknown character.
				throw new LexicalException("unknown character: " + ((char) c),
										   reader.getLineNumber() );
			}
			
			// if whitespace the token is set to null.
			if (token != null) tokenList.add(token);
			
		} while (c != -1);
		
		
		// allow reader to be garbage collected.
		reader = null;
		return tokenList;
	} // eom tokenize
	
	
	/** Scans an integer literal.<br>
	 *  integer literal ::= [0..0] ([0..9])*
	 *  @return The integer literal.
	 */
	protected String scanIntegerLiteral() throws IOException {
		reader.reset(); // reset to read first char of integer
		int c           = -1;
		int lineno      = reader.getLineNumber();
		StringBuffer il = new StringBuffer();
		do {
			reader.mark(PUSH_BACK_SIZE);
			c = reader.read();
			if (c != -1 && Character.isDigit((char) c)) {
				il.append((char) c);	
			} else {
				c = -1;
			}
		} while (c != -1);
		
		reader.reset(); // read one too far.
		return il.toString();
	} // eom scanIntLiteral
	
	
	/** Scans an identifier.<br>
 	 *	identifier ::= [a..Z] | '_' ([a..Z] | '_' | [0..9])*
	 *  returns the identifier.
	 */
	protected String scanIdentifier() throws IOException {
		reader.reset(); // reset to read first char of identifier
		int c           = -1;
		StringBuffer id = new StringBuffer();
		do {
			reader.mark(PUSH_BACK_SIZE);
			c = reader.read();
			if (Character.isJavaIdentifierPart((char) c)) {
				id.append((char) c);	
			} else {
				c = -1;
			}
		} while (c != -1);
		
		reader.reset(); // read one too far
		return id.toString();
	} // eom scanIdentifier
		
} // eoc DefaultLexicalAnalyzer
