import java.io.InputStream;
import java.util.List;

/** This class is responsible for syntactic analysis.
 *  The method parseProgram() recursivly calls other parseXXX() methods.
 *  Each of these methods returns an Abstract Syntax Tree (AST) for the construct it
 *  parses.
 *  The parsing process iterates through the list of tokens.
 *  @see ASTNode
 */ 
public class DefaultSyntaxAnalyzer implements SyntaxAnalyzer, ASTConstants, TOKConstants {
		
	/** The list of tokens (processed by LexicalAnalysis) */
	protected List tokens;
	
	/** The currently processed token. */
	protected TokenFactory.Token currentToken;
	
	/** The index of the currenttoken in the list of tokens. */
	protected int currentIndex = 0;
	
	/** Sets the list of tokens to parse. This method is synchronized. */
	public synchronized void setTokenList(List tokens) {
		this.tokens  = tokens;
		currentToken = (TokenFactory.Token) tokens.get(currentIndex);
	}
	
	/** Accepts the currentToken if its type is tokenType.
	 *  If it it accepted, currentToken is set to the next token.
	 *  @param tokenType The token type to accept.
	 *  @throws SyntaxException if currentToken.getType() != tokenType.
	 */
	void accept(int tokenType) throws SyntaxException {
		if (currentToken.getType() == tokenType) {
			acceptIt();	
		} else {
			// For real world error messages replace tokenType by human readable string
			throw new SyntaxException(currentToken, "Expected: " + tokenType + " here.");
		}
	}
	
	/** Accepts the currentToken. 
	 *  currentToken is set to then next token in the list of tokens. */
	void acceptIt() {
		currentIndex++;
		currentToken = (TokenFactory.Token) tokens.get(currentIndex);
	}
		
	/** Parses a program.
	   <pre>
	   PROGRAM ::= 'program' IDENTIFIER 
	               'declare' '{'
	                   DECLARATIONLIST 
				   '}' 
				   'run' '{' 
					   STATEMENTLIST
				   '}'</pre>
	*/
	public synchronized ASTNode parseProgram() throws SyntaxException {

		ASTNode programAST;
		
		if (currentToken.getType() == TokenFactory.TOK_PROGRAM) {
			programAST = new ASTNode(currentToken, AST_PROGRAM);
			acceptIt();
		} else {
			throw new SyntaxException(currentToken, "Expected PROGRAM here");
		}
		
		ASTNode programNameAST   = parseIdentifier();
		
		accept(TokenFactory.TOK_DECLARE);
		accept(TokenFactory.TOK_BLOCKBEGIN);
		ASTNode delarationAST    = parseDeclarationList();
		accept(TokenFactory.TOK_BLOCKEND);
		
		accept(TokenFactory.TOK_RUN);
		accept(TokenFactory.TOK_BLOCKBEGIN);
		ASTNode statementListAST = parseStatementList();
		accept(TokenFactory.TOK_BLOCKEND);
		
		if (currentToken.getType() != TokenFactory.TOK_EOT) {
			throw new SyntaxException(currentToken, 
			"Expected no more input after end of run block.");
		}

		programAST.add(programNameAST);
		programAST.add(delarationAST);
		programAST.add(statementListAST);
		
		return programAST;
	}
	
	
	/** Parses a Variable. 
	 *  <pre>VARIABLE ::= IDENTIFIER</pre>
	 */
	protected ASTNode parseVariable() throws SyntaxException {
		return parseIdentifier(); 	
	}
	
	
	/** Parses an Identifier. */
	protected ASTNode parseIdentifier() throws SyntaxException {
		ASTNode identifierAST;
		if (currentToken.getType() == TokenFactory.TOK_IDENTIFIER) {
			identifierAST = new ASTNode(currentToken, AST_IDENTIFIER);
			acceptIt();
		} else {			
			throw new SyntaxException(currentToken, "Expected IDENTIFIER here.");
		}
		return identifierAST;
	}
	
	
	/** Parses the declaration list.<br>
	 *  <pre>
	 *  DECLARATIONLIST ::=  DECLARATION (';' DECLARATION)* </pre> 
	 */
	protected ASTNode parseDeclarationList() throws SyntaxException {
		ASTNode declarationListAST = new ASTNode(currentToken, AST_DECLARATION_LIST);
		
		ASTNode declarationAST = parseDeclaration();
		declarationListAST.add(declarationAST);
		while (currentToken.getType() == TokenFactory.TOK_SEMICOLON) {
			acceptIt();
			declarationAST = parseDeclaration();
			declarationListAST.add(declarationAST);
		}
		
		return declarationListAST;
	} //eom parseDeclarationList();
	

	/** parses a declaration:
	 *  <pre>
	 *  DECLARATION ::= IDENTIFIER (',' IDENTIFIER)* ':' IDENTIFIER</pre>
	 */
	protected ASTNode parseDeclaration() throws SyntaxException {
		ASTNode declarationAST = new ASTNode(null, AST_DECLARATION);
		ASTNode identifierAST = parseIdentifier();
		declarationAST.add(identifierAST);
		while (currentToken.getType() == TokenFactory.TOK_COMMA) {
			acceptIt();
			identifierAST = parseIdentifier();
			declarationAST.add(identifierAST);
		}
		
		if (currentToken.getType() == TokenFactory.TOK_COLON) {
			declarationAST.setToken(currentToken);
			acceptIt();
		} else {
			throw new SyntaxException(currentToken, "Expected : here.");
		}

		if (currentToken.getType() == TokenFactory.TOK_IDENTIFIER) {
			identifierAST = parseIdentifier();
			declarationAST.add(identifierAST);
		} else {
			throw new SyntaxException(currentToken, "Expected typedenoter here.");
		}

		return declarationAST;
	} //eom parseDeclaration		
	

	/** Parses a statementlist:
	 *  <pre>
     *  STATEMENTLIST ::=  STATEMENT (';' STATEMENT)* </pre>
	 */
	protected ASTNode parseStatementList() throws SyntaxException {
		ASTNode statementListAST = new ASTNode (currentToken, AST_STATEMENT_LIST);
		
		ASTNode statementAST = parseStatement();
		statementListAST.add(statementAST);
		while (currentToken.getType() == TokenFactory.TOK_SEMICOLON) {
			acceptIt();
			statementAST = parseStatement();
			statementListAST.add(statementAST);
		}
		return statementListAST;
	} //eom parseStatementList
	
	
	/** Parses a statement:
	 *  <pre>
	 *  STATEMENT ::=  VARIABLE ':=' EXPRESSION
	 *				|  'if' EXPRESSION 'then' STATEMENT ('else' STATEMENT)+
	 *				|  'while' EXPRESSION 'do' STATEMENT
	 *              |  'for' EXPRESSION 'to' EXPRESSION 'do' STATEMENT
	 *              |  '{' STATEMENTLIST '}' </pre>
	 */
	protected ASTNode parseStatement() throws SyntaxException {
		ASTNode statementAST;
		if (currentToken.getType() == TokenFactory.TOK_IDENTIFIER) {
			statementAST = parseAssignment();
		} else
		if (currentToken.getType() == TokenFactory.TOK_IF) {
			statementAST = parseIfThenElse();
		} else		
		if (currentToken.getType() == TokenFactory.TOK_WHILE) {
			statementAST = parseWhileDo();
		} else		
		if (currentToken.getType() == TokenFactory.TOK_FOR) {
			statementAST = parseForToDo();
		} else		
		if (currentToken.getType() == TokenFactory.TOK_BLOCKBEGIN) {
			acceptIt();
			statementAST = parseStatementList();
			accept(TokenFactory.TOK_BLOCKEND);
		} else {
			throw new SyntaxException(currentToken, "Expected STATEMENT here.");
		}
		
		return statementAST;
	} // eom parseStatement
	
	
	/** Parses assignment:
	 *  <pre>
	 *  ASSIGNMENT ::= VARIABLE '=' EXPRESSION </pre>
	 */
	protected ASTNode parseAssignment() throws SyntaxException {
		ASTNode assignmentAST;
		ASTNode variableAST = parseVariable();
		if (currentToken.getType() == TokenFactory.TOK_ASSIGN) {
			assignmentAST = new ASTNode(currentToken, AST_ASSIGN);
			acceptIt();
		} else {
			throw new SyntaxException(currentToken, "Expected <var> = <expr> here.");
		}
		ASTNode expressionAST = parseExpression();
		
		assignmentAST.add(variableAST);
		assignmentAST.add(expressionAST);
		
		return assignmentAST;
	} // eom parseAssignment
	
	
	/** Parses an Expression.
	 *  <pre>
	 *  EXPRESSION ::= TERM ( OPLOW TERM)* </pre>
	 */
	protected ASTNode parseExpression() throws SyntaxException {
		ASTNode expressionAST; // = new ASTNode(currentToken, AST_EXPR);
		ASTNode operatorAST;
		ASTNode termAST;
		expressionAST = parseTerm();
		while (isLowOp(currentToken)) {
			operatorAST = parseBinaryOp();
			operatorAST.add(expressionAST);
			termAST = parseTerm();
			operatorAST.add(termAST);
			expressionAST = operatorAST;				
		}
		return expressionAST;
	} // eom parseExpression

	/** Tests if token is a low operator.
	 *  @param tok The token to test.
	 *  @return true if tok is a low operator. eg. TOK_PLUS
	 */
	public boolean isLowOp(TokenFactory.Token tok) {
		int tokType = tok.getType();
		return  tokType == TokenFactory.TOK_PLUS  ||
				tokType == TokenFactory.TOK_MINUS ||
				tokType == TokenFactory.TOK_EQ ||
				tokType == TokenFactory.TOK_NEQ ||
				tokType == TokenFactory.TOK_GT ||
				tokType == TokenFactory.TOK_LT ||
				tokType == TokenFactory.TOK_GTE ||
				tokType == TokenFactory.TOK_LTE;
	}
	
	/** Parses an Term.
	 *  <pre>
	 *  TERM ::= FACTOR ( OPLOW FACTOR )*</pre>
	 */
	protected ASTNode parseTerm() throws SyntaxException {
		ASTNode termAST;// = new ASTNode(currentToken, AST_TERM);
		ASTNode operatorAST;
		ASTNode factorAST;
		termAST = parseFactor();
		while (isHeightOp(currentToken)) {
			operatorAST = parseBinaryOp();
			operatorAST.add(termAST);
			factorAST = parseFactor();
			operatorAST.add(factorAST);
			termAST = operatorAST;			
		}
		return termAST;
	} // eom parseTerm
	

	/** Tests if token is a height operator.
	 *  @param tok The token to test.
	 *  @return true if tok is a height operator. eg. TOK_MUL
	 */
	public boolean isHeightOp(TokenFactory.Token tok) {
		int tokType = tok.getType();
		return  tokType == TokenFactory.TOK_MUL  ||
				tokType == TokenFactory.TOK_DIV;
	}
	
	
	/** Parses an Factor:
	 *  <pre>
	    FACTOR ::= '(' expr ')'
	 			 | INTLITERAL
				 | VARNAME
				 | BOOLLITERAL
				 | UNARYFACTOR </pre>
	 */
	protected ASTNode parseFactor() throws SyntaxException {
		ASTNode factorAST;
		if (currentToken.getType() == TokenFactory.TOK_OPEN_PAR) {
			acceptIt();
			factorAST = parseExpression();
			accept(TokenFactory.TOK_CLOSE_PAR);
		} else
		if (currentToken.getType() == TokenFactory.TOK_INT_LITERAL) {
			factorAST = parseIntLiteral();
		} else
		if (currentToken.getType() == TokenFactory.TOK_TRUE ||
		    currentToken.getType() == TokenFactory.TOK_FALSE) 
		{
			factorAST = parseBooleanLiteral();
		} else
		if (isUnaryOp(currentToken)) {
			factorAST = parseUnaryOp();
			ASTNode nextFactorAST = parseFactor();
			factorAST.add(nextFactorAST);
		} else 
		if (currentToken.getType() == TokenFactory.TOK_IDENTIFIER) {
			factorAST = parseVariable();
		} else {
			throw new SyntaxException(currentToken, "expected \'(\', literal or unary operator");	
		}
		return factorAST;
	} // eom parseFactor
	
	
	/** Tests if token is a unary operator.
	 *  @param tok The token to test.
	 *  @return true if tok is a unary operator. eg. TOK_PLUS, TOK_NOT
	 */
	public boolean isUnaryOp(TokenFactory.Token tok) {
		int tokType = tok.getType();
		return  tokType == TokenFactory.TOK_PLUS  ||
				tokType == TokenFactory.TOK_MINUS ||
				tokType == TokenFactory.TOK_NOT   ||
				tokType == TokenFactory.TOK_COMPLEMENT;
	}
	
	
	/** Parses an unary operator.
	 *  <pre>
	 *  UNARYOP ::= '+' | '-' | 'NOT' | '~' </pre>
	 */
	protected ASTNode parseUnaryOp() throws SyntaxException {
		ASTNode unaryNode = new ASTNode(currentToken, AST_UNARY_OP);
		if (isUnaryOp(currentToken)) {
			acceptIt();
			return unaryNode;
		} else {
			throw new SyntaxException(currentToken, "Expected unary operator here.");	
		}
	} // eom parseUnaryOp
	
	
	/** Tests if token is a binary operator.
	 *  @param tok The token to test.
	 *  @return true if tok is a binary operator. eg. TOK_PLUS
	 */
	public boolean isBinaryOp(TokenFactory.Token tok) {
		int tokType = tok.getType();
		return	tokType == TokenFactory.TOK_PLUS  ||
				tokType == TokenFactory.TOK_MINUS ||
				tokType == TokenFactory.TOK_MUL   ||
				tokType == TokenFactory.TOK_DIV   ||
				tokType == TokenFactory.TOK_AND   ||
				tokType == TokenFactory.TOK_OR    ||
				tokType == TokenFactory.TOK_EQ    ||
				tokType == TokenFactory.TOK_NEQ   ||
				tokType == TokenFactory.TOK_GT    ||
				tokType == TokenFactory.TOK_LT    ||
				tokType == TokenFactory.TOK_GTE   ||
				tokType == TokenFactory.TOK_LTE;  
	}
	
	
	/** Parses an binary operator.
	 */
	protected ASTNode parseBinaryOp() throws SyntaxException {
		ASTNode binaryNode = new ASTNode(currentToken, AST_BINARY_OP);
		if (isBinaryOp(currentToken)) {
				acceptIt();
				return binaryNode;
		} else {
			throw new SyntaxException(currentToken, "Expected binary operator here.");	
		}
	} // eom parseBinaryOp	
	
	
	/** Parses an boolean literal. */
	protected ASTNode parseBooleanLiteral() throws SyntaxException {
		ASTNode literalNode = new ASTNode(currentToken, AST_BOOLEAN_LITERAL);
		if (currentToken.getType() == TokenFactory.TOK_TRUE ||
		    currentToken.getType() == TokenFactory.TOK_FALSE)
		{
			acceptIt();
			return literalNode;
		} else {
			throw new SyntaxException(currentToken, "Expected boolean literal here.");	
		}
	} // eom parseBooleanLiteral
	
	
	/** Parses an int literal. */
	protected ASTNode parseIntLiteral() throws SyntaxException {
		ASTNode literalNode = new ASTNode(currentToken, AST_INT_LITERAL);
		if (currentToken.getType() == TokenFactory.TOK_INT_LITERAL) {
			acceptIt();
			return literalNode;
		} else {
			throw new SyntaxException(currentToken, "Expected integer literal here.");	
		}
	} // eom parseBooleanLiteral		
	
	
	/** Parses if then else.
	 *  Note: ( x )+ means zero or one time x
	 *  <pre>
	 *  IfThenElse ::=  'if' EXPRESSION 'then' STATEMENT ('else' STATEMENT)+ </pre>
	 */
	protected ASTNode parseIfThenElse() throws SyntaxException {
		ASTNode ifThenElseNode = new ASTNode(currentToken, AST_IF_THEN_ELSE);
		accept(TokenFactory.TOK_IF);
		ASTNode exprNode = parseExpression();
		accept(TokenFactory.TOK_THEN);
		ASTNode thenNode = parseStatement();
		
		ifThenElseNode.add(exprNode);
		ifThenElseNode.add(thenNode);
		
		if (currentToken.getType() == TokenFactory.TOK_ELSE) {
			acceptIt();
			ASTNode elseNode = parseStatement();
			ifThenElseNode.add(elseNode);
		}
		
		return ifThenElseNode;
	} // eom parseIfThenElse
	
	
	/** parses for to do.
	 *  <pre>
	    ForToDo ::= 'for' VARIABLE '=' EXPRESSION 'to' EXPRESSION 'do' STATEMENT </pre>
	 */
	protected ASTNode parseForToDo() throws SyntaxException {
		ASTNode forToDoNode = new ASTNode(currentToken, AST_FOR_TO_DO);
		accept(TokenFactory.TOK_FOR);
		ASTNode varNode = parseVariable();
		accept(TokenFactory.TOK_ASSIGN);
		ASTNode initNode = parseExpression();
		accept(TokenFactory.TOK_TO);
		ASTNode toNode = parseExpression();
		accept(TokenFactory.TOK_DO);
		ASTNode statementNode = parseStatement();
		
		forToDoNode.add(varNode);
		forToDoNode.add(initNode);
		forToDoNode.add(toNode);
		forToDoNode.add(statementNode);
		
		return forToDoNode;
	} // eom parseForToDo
	
	
	/** Parses while do
	 *  <pre>
	    WhileDo ::= 'while' EXPRESSION 'do' STATEMENT </pre>
	 */
	protected ASTNode parseWhileDo() throws SyntaxException {
		ASTNode whileDoNode = new ASTNode(currentToken, AST_WHILE_DO);
		accept(TokenFactory.TOK_WHILE);
		ASTNode exprNode = parseExpression();
		accept(TokenFactory.TOK_DO);
		ASTNode statementNode = parseStatement();
		
		whileDoNode.add(exprNode);
		whileDoNode.add(statementNode);
		
		return whileDoNode;
	} // eom parseWhileDo
	
} // eoc DefaultSyntaxAnalyzer
